#ifndef __tinyvector__
#define __tinyvector__

// Bereitstellung eines simplen Containers, der eine Datenfolge
// im dynamischen Array verwaltet.
// Hauptaufgabe: sparse-Darstellung der Faktorenindices der
// statischen Faktorbasis (fr qsieve).
// (C) 1999 by Thorsten Reinecke

template <class Datatype> class CTinyVector
{
public:
  int akt_size; // aktuelle Gre des Vektors
  int max_size; // maximale Gre des Vektors
  Datatype* v; // dynamisches Feld
  
  inline CTinyVector(const int MaxSize = 20) // Konstruktur
    {
      max_size=MaxSize;
      akt_size=0;
      v = new Datatype [MaxSize];
    }
  inline CTinyVector(const Datatype* Feld, const int Size)
    // Konstruktur: mit vorgegebenem Feld initialisieren
    {
      akt_size=max_size=Size;
      v = new Datatype [Size];
      for(int i=0; i<Size; i++) v[i]=Feld[i];
    }
  inline ~CTinyVector(void) // Destruktor
    {
      delete [] v;
    }
  inline void reset(void)
    {
      akt_size=0;
    }
  inline void copy_from(const Datatype* Feld, const int Size)
    {
      if (Size>max_size) { delete [] v; v = new Datatype [Size]; max_size=Size; }
      akt_size=Size;
      for(int i=0; i<Size; i++) v[i]=Feld[i];
    }
  inline void optisized_copy_from(const Datatype* Feld, const int Size)
   {
      if (Size!=max_size) { delete [] v; v = new Datatype [Size]; }
      akt_size=max_size=Size;
      for(int i=0; i<Size; i++) v[i]=Feld[i];
   }
  inline void resize(const int new_size)
    {
      if (new_size<akt_size)
	{ cerr << "Tiny-Vector: Resizing not possible!"; exit(1); }
      if (new_size==max_size) return; // nothing to do...
      Datatype* vnew = new Datatype [new_size];
      for (int i=0; i<akt_size; i++) vnew[i]=v[i];
      delete [] v; v=vnew; max_size=new_size;
    }
  inline void optisize(void) // Gre des dyn. Array optimieren
    {
      resize(akt_size);
    }
  inline void append(const Datatype value)
    {
      if (akt_size>=max_size) 
	{
	  cerr << "TinyVector: Index " << akt_size
	       << " is out Of Range!"
	       << " Auto-Resizing by 5" << endl;
	  resize(akt_size+5);
	}
      if (!empty() && value<=v[akt_size-1])
	{ cerr << "TinyVector: unsortiertes append!" << endl; } 
      v[akt_size++]=value; 
    }
  inline void fast_append(const Datatype value)
    {
      v[akt_size++]=value;
    }
  inline int size(void) const
    {
      return akt_size;
    }
  inline bool empty(void) const
    {
      return (akt_size==0);
    }
  inline Datatype last(void) const
    {
      if (!empty()) return v[akt_size-1];
      else
	{
	  cerr << "TinyVector: no last Element!" << endl;
	  return -1;
	}
    }
};

#endif
