// Implementation of elliptic curves method (ecm) V0.8b
// written by Thorsten Reinecke, 23.06.99
// last change: 04.07.99
// scheint zu laufen, wenn auch recht langsam...

//#define DEBUGMODE


class elliptic_curves
{
 private:
  mpz_t a,b; // auer in go nur lesend verwendbar!
  mpz_t h,k,m; // allgemein verwendbar (zwischen Aufrufen vernderbar)!
  mpz_t x3,y3; // nur in add, mul2 verwendbar!
  mpz_t xh_mul, yh_mul; // nur in mul verwendbar!
 public:
  inline elliptic_curves() /* Konstruktor */
    {
      mpz_init(a); mpz_init(b);
      mpz_init(h); mpz_init(k); mpz_init(m);
      mpz_init(x3); mpz_init(y3);
      mpz_init(xh_mul); mpz_init(yh_mul);
    }
  inline ~elliptic_curves() /* Destruktor */
    {
      mpz_clear(a); mpz_clear(b);
      mpz_clear(h); mpz_clear(k); mpz_clear(m);
      mpz_clear(x3); mpz_clear(y3); 
      mpz_clear(xh_mul); mpz_clear(yh_mul);
    }
 private:
  void check_curve(const mpz_t x, const mpz_t y);
  void factor_found(const mpz_t &k);
  bool add(mpz_t &xr, mpz_t &yr,
           const mpz_t x1, const mpz_t y1,
           const mpz_t x2, const mpz_t y2);
  bool mul2(mpz_t &xr, mpz_t &yr, const mpz_t x1, const mpz_t y1);
  bool mul(mpz_t &xr, mpz_t &yr,
           const mpz_t x1, const mpz_t y1,
           unsigned int K);
 public:
  void go(const int phase1, const int phase2);

};     


void elliptic_curves::check_curve(const mpz_t x, const mpz_t y)
{
  // prft, ob der Punkt auf der Kurve liegt, sonst: Fehlermeldung & Abbruch
  mpz_t h,x1,y1;
  mpz_init(x1); mpz_init(y1); mpz_init(h);
  mpz_powm_ui(x1,x,3,n); mpz_mul(h,a,x); mpz_add(x1,x1,h); mpz_add(x1,x1,b);
  mpz_mod(x1,x1,n);
  mpz_powm_ui(y1,y1,2,n);
  int fl=mpz_cmp(x1,y1);
  mpz_clear(x1); mpz_clear(y1); mpz_clear(h);
  if (fl!=0)
   {
     cerr << "Geprfter Punkt liegt nicht auf der Kurve!" << endl;
     exit(1);
   }
}


void elliptic_curves::factor_found(const mpz_t &k)
{
  cout << endl << "Faktor gefunden in elliptic curve" << endl;
  if (mpz_cmp(k,n)==0) // gefundener Faktor = n ?
   {
     cout << "Gefundene Faktorisierung ist trivial. Breche ab." << endl;
     return; // trivialer Teiler -> Verfahren abbrechen
   }
  if (mpz_probab_prime_p(k,probab_prime_checks))
    {
      cout << k << " ist Faktor." << endl;
      Factorization_to_file << MAL() << k << " [ecm]" << flush;
    }
  else
    {
      cout << k << " ist zusammengesetzter Faktor." << endl;
      Factorization_to_file << MAL() << k << " [ecm] [compound]" << flush;
    }
  mpz_div(n,n,k);
}

bool elliptic_curves::mul2(mpz_t &xr, mpz_t &yr, const mpz_t x1, const mpz_t y1)
{
  //cout << "";
  mpz_mul_ui(k,y1,2); mpz_mod(k,k,n);
  if (!mpz_invert(k,k,n))
    {
      mpz_mul_ui(k,y1,2); mpz_mod(k,k,n); mpz_gcd(k,k,n);
      factor_found(k);
      return true; // mit dieser Kurve sind wir fertig
    }
  mpz_mul(m,x1,x1); mpz_mod(m,m,n); mpz_mul_ui(m,m,3); mpz_add(m,m,a);
  mpz_mul(m,m,k); mpz_mod(m,m,n);
  mpz_mul(x3,m,m); mpz_mul_ui(k,x1,2); mpz_sub(x3,x3,k); mpz_mod(x3,x3,n);
  mpz_sub(y3,x1,x3); mpz_mul(y3,y3,m); mpz_sub(y3,y3,y1); mpz_mod(yr,y3,n);
  mpz_set(xr,x3);
  return false; // kein Faktor gefunden
}

bool elliptic_curves::add(mpz_t &xr, mpz_t &yr,
                          const mpz_t x1, const mpz_t y1,
                          const mpz_t x2, const mpz_t y2)
{
  if (mpz_cmp(x1,x2)) // x1!=x2
   {
     //cout << "'";
     mpz_sub(k,x2,x1); mpz_mod(k,k,n);
     if (!mpz_invert(k,k,n))
      {
        mpz_sub(k,x2,x1); mpz_mod(k,k,n); mpz_gcd(k,k,n);
        factor_found(k);
        return true; // mit dieser Kurve sind wir fertig
      }
     mpz_sub(m,y2,y1); mpz_mul(m,m,k); mpz_mod(m,m,n);
     mpz_mul(x3,m,m); mpz_sub(x3,x3,x1); mpz_sub(x3,x3,x2); mpz_mod(x3,x3,n);
     mpz_sub(y3,x1,x3); mpz_mul(y3,y3,m); mpz_sub(y3,y3,y1); mpz_mod(y3,y3,n);
   }
  else // x1==x2
   {
     if (mpz_cmp(y1,y2)==0) // y1==y2
       {
         //cout << ".";
         mpz_mul_ui(k,y1,2); mpz_mod(k,k,n);
         if (!mpz_invert(k,k,n))
          {
            mpz_mul_ui(k,y1,2); mpz_mod(k,k,n); mpz_gcd(k,k,n);
            factor_found(k);
            return true; // mit dieser Kurve sind wir fertig
          }
	 mpz_mul(m,x1,x1); mpz_mod(m,m,n); mpz_mul_ui(m,m,3); mpz_add(m,m,a);
         mpz_mul(m,m,k); mpz_mod(m,m,n);
         mpz_mul(x3,m,m); mpz_mul_ui(k,x1,2); mpz_sub(x3,x3,k); mpz_mod(x3,x3,n);
	 mpz_sub(y3,x1,x3); mpz_mul(y3,y3,m); mpz_sub(y3,y3,y1); mpz_mod(y3,y3,n);
       }
      else // y1!=y2
       {
         cout << "Elliptic curves: Nicht implementierter Sonderfall" << endl;
         return true;
       } 
   }

  mpz_set(xr,x3); mpz_set(yr,y3);
  return false; // kein Faktor gefunden
}

bool elliptic_curves::mul(mpz_t &xr, mpz_t &yr,
                        const mpz_t x1, const mpz_t y1,
                        unsigned int K)
{
  mpz_set(xh_mul,x1); mpz_set(yh_mul,y1);

  //kleines Problem: Addition mit neutralem Element mit add nicht mglich!
  //also wird folgendermaen verfahren:
#ifdef DEBUGMODE
  if (K==0) { cerr << "Multiplikation mit Null nicht erlaubt!"; exit(1); }
#endif
  while ((K&1)==0)
   {
     K>>=1;
     if (mul2(xh_mul,yh_mul,xh_mul,yh_mul)) return true; // Faktor gefunden
   }
  // nun gilt (K&1)==1
  mpz_set(xr,xh_mul); mpz_set(yr,yh_mul);
  while (K>>=1)
   {
     if (mul2(xh_mul,yh_mul,xh_mul,yh_mul)) return true; // Faktor gefunden
     if (K&1) if (add(xr,yr,xr,yr,xh_mul,yh_mul)) return true; // Faktor gefunden
   }
  return false; // kein Faktor gefunden
}

void elliptic_curves::go(const int phase1, const int phase2)
{
  mpz_t x,y,delta, saved_x, saved_y, xh,yh;
  mpz_init(x); mpz_init(y); mpz_init(delta);
  mpz_init(saved_x); mpz_init(saved_y);
  mpz_init(xh); mpz_init(yh);

  
  mpz_set_ui(x,rand()); mpz_mod(x,x,n);
  mpz_set_ui(y,rand()); mpz_mod(y,y,n);
  mpz_set_ui(a,rand()); mpz_mod(a,a,n);

  cout << "Trying Elliptic Curve Method with " << endl;
  cout << "x=" << x << endl;
  cout << "y=" << y << endl;
  cout << "a=" << a << endl;

  mpz_mul(b,y,y); mpz_powm_ui(h,x,3,n); mpz_sub(b,b,h);
  mpz_mul(h,a,x); mpz_sub(b,b,h); mpz_mod(b,b,n);
  cout << "b=" << b << endl;

  mpz_powm_ui(h,a,3,n); mpz_mul_ui(h,h,4);
  mpz_powm_ui(delta,b,2,n); mpz_mul_ui(delta,delta,27);
  mpz_add(delta,delta,a); mpz_mod(delta,delta,n);

  if (mpz_cmp_ui(delta,0)==0)
   {
     cerr << "Elliptic curves: Sonderfall eingetreten..." << endl;
     // hier mu einfach eine neue Kurve gewhlt werden...
     go(phase1,phase2); // was hiermit geschieht ;)
     return;
   }

  mpz_gcd(h,delta,n);
  if (mpz_cmp_ui(h,1)!=0)
   {
     // zufllig gleich mit der Kurve einen Faktor entdeckt...
     factor_found(h);
     return;
   }
  
  int i,j,k,d, z=1;
  // Phase 0: kleine Primzahlen (mit hohen Primzahlpotenzen)
  cout << "Elliptic Curve Phase 0" << endl;
  for (j=1; j<=50; j++) if (mul2(x,y,x,y)) goto fertig; // 2er-Potenzen
  for (j=1; j<=25; j++) if (mul(x,y,x,y,3)) goto fertig; // 3er-Potenzen
  d=2;
  for (i=3; i<500; i++) if (is_prime(i))
   {
     i+=d; d=6-d;
     j=phase2; k=1;
     while (j>=i) { k*=i; j/=i; }
     if (mul(x,y,x,y,k)) goto fertig;
     if (mul(x,y,x,y,k)) goto fertig; // ruhig zweimal...
   }

  // Phase 1: Primzahlen mit (niedrigen) Potenzen
  i=i-(i%6)-1-6; d=2; // i mit Eigenschaft i=5 (mod 6) fr schnellere Primzahlermittlung
  cout << "Elliptic Curve Phase 1" << endl;
  while (i<=phase1)
    {
      do { i+=d; d=6-d; } while (i%5==0 || i%7==0 || i%11==0 || i%13==0 ||
                                 i%17==0 || i%19==0 || i%23==0 || i%29==0 ||
                  		 i%31==0 || i%37==0 || i%41==0 || i%43==0 ||
           		         i%47==0 || i%53==0 || i%59==0 || i%61==0 ||
				 powmod(2,i-1,i)!=1);
        // obige do-Schleife soll mglichst schnell viele zusammengesetzte
        // Zahlen ausschlieen, powmod(..) = Fermat-Test
      if (--z==0)
       { // Status-Ausgabe
         z=100;
         cout << "Phase1: " << i << "/" << phase1 << "\r" << flush;
       }
      j=phase2; k=1;
      while (j>=i) { k*=i; j/=i; }
      if (mul(x,y,x,y,k)) goto fertig;
    }
  cout << "Elliptic Curve Phase 1 bis " << i << endl;

  {
   // Phase 2: isolierte (Prim-)zahl: Kurve mu bis auf diese Zahl vorher vollstndig zerlegbar gewesen sein.
   cout << "Elliptic Curve Phase 2" << endl;
   mpz_set(saved_x,x); mpz_set(saved_y,y); // Basis fr Phase2

   // precomputing most common used differences:
   const int pre_exp_size = 100;
   mpz_t bfact[pre_exp_size][2];

   if (mul2(x,y,saved_x,saved_y)) goto fertig;
   mpz_set(xh,x); mpz_set(yh,y);
   for (int j=2; j<pre_exp_size; j+=2)
    {
      mpz_init_set(bfact[j][0],x); mpz_init_set(bfact[j][1],y);
      if (add(x,y,x,y,xh,yh)) goto fertig;
    }  

   while (!probab_prime(i)) { i+=d; d=6-d; }
   if (mul(x,y,saved_x,saved_y,i)) goto fertig_phase2;
  
   while (i<=phase2) 
    { 
      int runden=2000;
      while (runden>0)
	{
	  int i_prev=i;
          do { i+=d; d=6-d; } while (i%5==0 || i%7==0 || i%11==0 || i%13==0 ||
                                     i%17==0 || i%19==0 || i%23==0 || i%29==0 ||
	           		     i%31==0 || i%37==0 || i%41==0 || i%43==0 ||
	           		     i%47==0 || i%53==0 || i%59==0 || i%61==0);
				     // || powmod(2,i-1,i)!=1);
	 // obige do-Schleife soll mglichst schnell viele zusammengesetzte
         // Zahlen ausschlieen, (Fermat-Test hier zu teuer!) 
          if (i-i_prev<pre_exp_size)
           {
             if (add(x,y,x,y,bfact[i-i_prev][0],bfact[i-i_prev][1])) goto fertig_phase2;
           }
          else
           { // not a precomputed difference
	     cerr << "not a precomputed exponent " << i-i_prev << endl;
             if (mul(xh,yh,saved_x,saved_y,i-i_prev)) goto fertig_phase2;
             if (add(x,y,x,y,xh,yh)) goto fertig_phase2;
           }
	  --runden;
	}
      cout << "Phase2: " << i << "/" << phase2 << "\r" << flush;
    }

 fertig_phase2:
  cout << "Elliptic Curve Phase 2 bis " << i << endl;
  for (int j=2; j<pre_exp_size; j+=2)
    {
      // freeing memory for precomputed differences
      mpz_clear(bfact[j][0]); mpz_clear(bfact[j][1]);
    }
  }


fertig:
  mpz_clear(x); mpz_clear(y); mpz_clear(delta);
  mpz_clear(saved_x); mpz_clear(saved_y);
  mpz_clear(xh); mpz_clear(yh);
}
